@@ -20,10 +20,25 @@ type mat = float64_mat
 
 (* Conversion of OCaml-values to S-expressions *)
 
-(* Some basic experiments indicate that %.20G is enough to round-trip
-   a float through the sexp-converter, (although that was done long ago,
-   and there's no real guarantee) *)
-let default_string_of_float = ref (fun n -> sprintf "%.20G" n)
+external format_float : string -> float -> string = "caml_format_float"
+
+(* '%.17g' is guaranteed to be round-trippable.
+
+   '%.15g' will be round-trippable and not have noise at the last digit or two for a float
+   which was converted from a decimal (string) with <= 15 significant digits.  So it's
+   worth trying first to avoid things like "3.1400000000000001".
+
+   See comment above [to_string_round_trippable] in [base/core/kernel/lib/float.ml] for
+   detailed explanation and examples. *)
+let default_string_of_float =
+  ref (fun x ->
+    let y = format_float "%.15G" x in
+    if float_of_string y = x then
+      y
+    else
+      format_float "%.17G" x)
+;;
+
 let read_old_option_format = ref true
 let write_old_option_format = ref true
 
@@ -38,10 +53,6 @@ let sexp_of_float n = Atom (!default_string_of_float n)
 let sexp_of_int32 n = Atom (Int32.to_string n)
 let sexp_of_int64 n = Atom (Int64.to_string n)
 let sexp_of_nativeint n = Atom (Nativeint.to_string n)
-let sexp_of_big_int n = Atom (Big_int.string_of_big_int n)
-let sexp_of_nat n = Atom (Nat.string_of_nat n)
-let sexp_of_num n = Atom (Num.string_of_num n)
-let sexp_of_ratio n = Atom (Ratio.string_of_ratio n)
 let sexp_of_ref sexp_of__a rf = sexp_of__a !rf
 let sexp_of_lazy_t sexp_of__a lv = sexp_of__a (Lazy.force lv)
 
@@ -113,6 +124,7 @@ let string_of__of__sexp_of to_sexp x = Sexp.to_string (to_sexp x)
 
 (* Exception converter registration and lookup *)
 
+
 module Exn_converter = struct
   type t = int64
 
@@ -161,12 +173,66 @@ module Exn_converter = struct
 
   (* Fast and automatic exception registration *)
 
+  let max_exn_tags = ref 20
+
+  let set_max_exn_tags n =
+    if n < 1 then
+      failwith "Sexplib.Conv.Exn_converter.set_max_exn_tags: n < 1"
+    else max_exn_tags := n
+
+  let get_max_exn_tags () = !max_exn_tags
+
   module Int = struct
     type t = int
 
     let compare t1 t2 = compare (t1 : int) t2
   end
 
+IFDEF OCAML_4_02 THEN
+
+  (* This is not exposed in [Obj] *)
+  let extension_slot x =
+    let x = Obj.repr x in
+    if Obj.tag x = Obj.object_tag then x else Obj.field x 0
+  ;;
+
+  module Exn_ids = Map.Make (Int)
+
+  let exn_id_map : (exn -> Sexp.t) Exn_ids.t ref = ref Exn_ids.empty
+
+  let rec clean_up_handler slot =
+    let id = Obj.extension_id slot in
+    let old_exn_id_map = !exn_id_map in
+    let new_exn_id_map = Exn_ids.remove id old_exn_id_map in
+    (* This trick avoids mutexes and should be fairly efficient *)
+    if !exn_id_map != old_exn_id_map then
+      clean_up_handler slot
+    else
+      exn_id_map := new_exn_id_map
+
+  let add_auto ?(finalise = true) exn sexp_of_exn =
+    let id = Obj.extension_id exn in
+    let rec loop () =
+      let old_exn_id_map = !exn_id_map in
+      let new_exn_id_map = Exn_ids.add id sexp_of_exn old_exn_id_map in
+      (* This trick avoids mutexes and should be fairly efficient *)
+      if !exn_id_map != old_exn_id_map then
+        loop ()
+      else begin
+        exn_id_map := new_exn_id_map;
+        if finalise then Gc.finalise clean_up_handler (extension_slot exn)
+      end
+    in
+    loop ()
+
+  let find_auto exn =
+    let id = Obj.extension_id exn in
+    match try Some (Exn_ids.find id !exn_id_map) with Not_found -> None with
+    | None -> None
+    | Some sexp_of_exn -> Some (sexp_of_exn exn)
+
+ELSE
+
   module Addrs = Map.Make (Int)
 
   type weak_repr = (Obj.t Weak.t * (exn -> Sexp.t)) Ids.t
@@ -201,15 +267,6 @@ module Exn_converter = struct
 
   exception Found_sexp of Sexp.t
 
-  let max_exn_tags = ref 20
-
-  let set_max_exn_tags n =
-    if n < 1 then
-      failwith "Sexplib.Conv.Exn_converter.set_max_exn_tags: n < 1"
-    else max_exn_tags := n
-
-  let get_max_exn_tags () = !max_exn_tags
-
   let add_auto ?(finalise = true) exn sexp_of_exn =
     let exn_tag = get_exn_tag exn in
     let addr = get_exn_tag_str_addr exn_tag in
@@ -255,6 +312,8 @@ module Exn_converter = struct
           Ids.iter act exn_handler_map;
           None
         with Found_sexp sexp -> Some sexp
+
+ENDIF
 end
 
 let sexp_of_exn_opt exn =
@@ -336,34 +395,6 @@ let nativeint_of_sexp sexp = match sexp with
         of_sexp_error ("nativeint_of_sexp: " ^ exn_to_string exc) sexp)
   | List _ -> of_sexp_error "nativeint_of_sexp: atom needed" sexp
 
-let big_int_of_sexp sexp = match sexp with
-  | Atom str ->
-      (try Big_int.big_int_of_string str
-      with exc ->
-        of_sexp_error ("big_int_of_sexp: " ^ exn_to_string exc) sexp)
-  | List _ -> of_sexp_error "big_int_of_sexp: atom needed" sexp
-
-let nat_of_sexp sexp = match sexp with
-  | Atom str ->
-      (try Nat.nat_of_string str
-      with exc ->
-        of_sexp_error ("nat_of_sexp: " ^ exn_to_string exc) sexp)
-  | List _ -> of_sexp_error "nat_of_sexp: atom needed" sexp
-
-let num_of_sexp sexp = match sexp with
-  | Atom str ->
-      (try Num.num_of_string str
-      with exc ->
-        of_sexp_error ("num_of_sexp: " ^ exn_to_string exc) sexp)
-  | List _ -> of_sexp_error "num_of_sexp: atom needed" sexp
-
-let ratio_of_sexp sexp = match sexp with
-  | Atom str ->
-      (try Ratio.ratio_of_string str
-      with exc ->
-        of_sexp_error ("ratio_of_sexp: " ^ exn_to_string exc) sexp)
-  | List _ -> of_sexp_error "ratio_of_sexp: atom needed" sexp
-
 let ref_of_sexp a__of_sexp sexp = ref (a__of_sexp sexp)
 let lazy_t_of_sexp a__of_sexp sexp = Lazy.lazy_from_val (a__of_sexp sexp)
 
@@ -509,7 +540,7 @@ let of_string__of__of_sexp of_sexp s =
 (* Registering default exception printers *)
 
 let get_flc_error name (file, line, chr) =
-  List [Atom name; Atom file; sexp_of_int line; sexp_of_int chr]
+  Atom (sprintf "%s %s:%d:%d" name file line chr)
 
 let () =
   List.iter
@@ -605,13 +636,6 @@ let () =
         (function
         | Sys.Break -> Atom "Sys.Break"
         | _ -> assert false)
-      );(
-        Unix.Unix_error (Unix.E2BIG, "", ""),
-        (function
-        | Unix.Unix_error (err, loc, arg) ->
-            let err_str = Unix.error_message err in
-            List [Atom "Unix.Unix_error"; Atom err_str; Atom loc; Atom arg]
-        | _ -> assert false)
       );(
         Of_sexp_error (Exit, Atom ""),
         (function
