@@ -14,6 +14,16 @@ module Gen = Pa_type_conv.Gen
 
 (* Utility functions *)
 
+let rec go_through_private_and_manifest_types = function
+  | <:ctyp< private $tp$ >> -> (true, tp)
+  | <:ctyp< $_$ == $tp$ >>  -> go_through_private_and_manifest_types tp
+  | tp                      -> (false, tp)
+
+let replace_variables_by_underscores =
+  (Ast.map_ctyp (function
+    | <:ctyp@loc< '$_$ >> -> <:ctyp@loc< _ >>
+    | ctyp -> ctyp))#ctyp
+
 let mk_rev_bindings loc fps =
   let coll (i, bindings, patts, vars) fp =
     let name = "v" ^ string_of_int i in
@@ -77,14 +87,29 @@ let type_app base_type types =
    non-regular types, such as [ type t 'a = [ A of 'a | B of t 'a 'a ] ]. *)
 let mk_poly_type make_mono_type loc type_name type_params =
   let type_params = List.map type_params ~f:Gen.drop_variance_annotations in
+  let unquantified_type = make_mono_type <:ctyp@loc< $lid:type_name$ >> type_params in
+  let loc = Ast.loc_of_ctyp unquantified_type in
   match type_params with
-  | [] -> None
+  | [] ->
+    <:ctyp@loc< $unquantified_type$ >>
   | first :: rest ->
-    let unquantified_type = make_mono_type <:ctyp@loc< $lid:type_name$ >> type_params in
-    let loc = Ast.loc_of_ctyp unquantified_type in
     (* It's confusing that you need an application between the '!' and the '.'. I would
        have expected a list. *)
-    Some <:ctyp@loc< ! $type_app first rest$ . $unquantified_type$ >>
+    <:ctyp@loc< ! $type_app first rest$ . $unquantified_type$ >>
+
+(* This transformation is sensible only when the gadt syntax is used to describe regular
+   variants or existential types but not when the return type is constrained like in [type
+   'a t = Int : int t]. In this last case, the generated code is not going to compile
+   because [sexp_of_t] would have type [('a -> Sexp.t) -> int t -> Sexp.t] and there is
+   type constraint ['a. ('a -> Sexp.t) -> 'a t -> Sexp.t]. *)
+let regular_variant_of_gadt_syntax = function
+  | <:ctyp@loc< $uid:cnstr$ : $args$ -> $return_type$ >> ->
+    ignore return_type;
+    <:ctyp@loc< $uid:cnstr$ of $args$ >>
+  | <:ctyp@loc< $uid:cnstr$ : $return_type$ >> ->
+    ignore return_type;
+    <:ctyp@loc< $uid:cnstr$ >>
+  | ctyp -> ctyp
 
 (* Generators for S-expressions *)
 
@@ -287,8 +312,10 @@ module Generate_sexp_of = struct
       | <:ctyp@loc< $tp1$ | $tp2$ >> ->
           <:match_case@loc< $loop tp1$ | $loop tp2$ >>
       | <:ctyp@loc< `$cnstr$ >> ->
-          <:match_case@loc< `$cnstr$ -> Sexplib.Sexp.Atom $str:cnstr$ >>
+          let str = Pa_type_conv.Gen.regular_constr_of_revised_constr cnstr in
+          <:match_case@loc< `$cnstr$ -> Sexplib.Sexp.Atom $str:str$ >>
       | <:ctyp@loc< `$cnstr$ of sexp_list $tp$>> ->
+        let str = Pa_type_conv.Gen.regular_constr_of_revised_constr cnstr in
         let cnv_expr =
           match sexp_of_type tp with
           | `Fun fun_expr -> <:expr@loc< $fun_expr$ >>
@@ -298,13 +325,14 @@ module Generate_sexp_of = struct
         <:match_case@loc<
           `$cnstr$ l ->
              Sexplib.Sexp.List
-               [ Sexplib.Sexp.Atom $str:cnstr$ ::
+               [ Sexplib.Sexp.Atom $str:str$ ::
                    Sexplib.Conv.list_map $cnv_expr$ l]
         >>
       | <:ctyp@loc< `$cnstr$ of $tps$ >> ->
+          let str = Pa_type_conv.Gen.regular_constr_of_revised_constr cnstr in
           let fps = List.map ~f:sexp_of_type (Ast.list_of_ctyp tps []) in
           let bindings, patts, vars = mk_bindings loc fps in
-          let cnstr_expr = <:expr@loc< Sexplib.Sexp.Atom $str:cnstr$ >> in
+          let cnstr_expr = <:expr@loc< Sexplib.Sexp.Atom $str:str$ >> in
           let expr =
             <:expr@loc<
               let $bindings$ in
@@ -354,12 +382,15 @@ module Generate_sexp_of = struct
 
   (* Conversion of sum types *)
 
-  let rec branch_sum = function
+  let rec branch_sum ctyp =
+    match regular_variant_of_gadt_syntax ctyp with
     | <:ctyp@loc< $tp1$ | $tp2$ >> ->
         <:match_case@loc< $branch_sum tp1$ | $branch_sum tp2$ >>
     | <:ctyp@loc< $uid:cnstr$ >> ->
-        <:match_case@loc< $uid:cnstr$ -> Sexplib.Sexp.Atom $str:cnstr$ >>
+        let str = Pa_type_conv.Gen.regular_constr_of_revised_constr cnstr in
+        <:match_case@loc< $uid:cnstr$ -> Sexplib.Sexp.Atom $str:str$ >>
     | <:ctyp@loc< $uid:cnstr$ of sexp_list $tp$>> ->
+        let str = Pa_type_conv.Gen.regular_constr_of_revised_constr cnstr in
         let cnv_expr =
           match sexp_of_type tp with
           | `Fun fun_expr -> <:expr@loc< $fun_expr$ >>
@@ -369,12 +400,13 @@ module Generate_sexp_of = struct
         <:match_case@loc<
           $uid:cnstr$ l ->
              Sexplib.Sexp.List
-               [Sexplib.Sexp.Atom $str:cnstr$ ::
+               [Sexplib.Sexp.Atom $str:str$ ::
                    Sexplib.Conv.list_map $cnv_expr$ l]
         >>
     | <:ctyp@loc< $uid:cnstr$ of $tps$ >> ->
+        let str = Pa_type_conv.Gen.regular_constr_of_revised_constr cnstr in
         let fps = List.map ~f:sexp_of_type (Ast.list_of_ctyp tps []) in
-        let cnstr_expr = <:expr@loc< Sexplib.Sexp.Atom $str:cnstr$ >> in
+        let cnstr_expr = <:expr@loc< Sexplib.Sexp.Atom $str:str$ >> in
         let bindings, patts, vars = mk_bindings loc fps in
         let patt =
           match patts with
@@ -386,8 +418,6 @@ module Generate_sexp_of = struct
             let $bindings$ in
             Sexplib.Sexp.List $Gen.mk_expr_lst loc (cnstr_expr :: vars)$
         >>
-    | <:ctyp< $_$ : $_$ >> as tp -> Gen.error tp ~fn:"branch_sum"
-      ~msg:"GADTs are not supported by sexplib"
     | tp -> Gen.unknown_type tp "branch_sum"
 
   let sexp_of_sum alts = `Match (branch_sum alts)
@@ -540,26 +570,37 @@ module Generate_sexp_of = struct
   (* Empty type *)
   let sexp_of_nil loc = `Fun <:expr@loc< fun _v -> assert False >>
 
-
   (* Generate code from type definitions *)
 
   let sexp_of_td loc type_name tps rhs =
+    let is_private, rhs = go_through_private_and_manifest_types rhs in
     let body =
-      let rec loop tp =
-        Gen.switch_tp_def tp
-          ~alias:(fun (_ : Loc.t) tp -> sexp_of_type tp)
-          ~sum:(fun (_ : Loc.t) tp -> sexp_of_sum tp)
-          ~record:(fun (_ : Loc.t) tp -> sexp_of_record tp)
-          ~variants:(fun (_ : Loc.t) tp -> sexp_of_variant tp)
-          ~mani:(fun (_ : Loc.t) _tp1 tp2 -> loop tp2)
-          ~nil:sexp_of_nil
+      let is_private_alias, body =
+        Gen.switch_tp_def rhs
+          ~alias:   (fun (_ : Loc.t) tp  -> (is_private, sexp_of_type    tp))
+          ~sum:     (fun (_ : Loc.t) tp  -> (false     , sexp_of_sum     tp))
+          ~record:  (fun (_ : Loc.t) tp  -> (false     , sexp_of_record  tp))
+          ~variants:(fun (_ : Loc.t) tp  -> (false     , sexp_of_variant tp))
+          ~mani:    (fun (_ : Loc.t) _ _ -> assert false                    )
+          ~nil:     (fun loc             -> (false     , sexp_of_nil loc   ))
       in
-      match loop rhs with
-      | `Fun fun_expr ->
-          (* Prevent violation of value restriction and problems with
-             recursive types by eta-expanding function definitions *)
+      if is_private_alias then
+        (* Replace all type variable by _ to avoid generalization problems *)
+        let ty_src = mk_full_type loc type_name tps in
+        let ty_dst = replace_variables_by_underscores rhs in
+        let coercion = <:expr@loc< (v : $ty_src$ :> $ty_dst$) >> in
+        match body with
+        | `Fun fun_expr ->
+          <:expr@loc< fun [ v -> $fun_expr$ $coercion$ ] >>
+        | `Match matchings ->
+          <:expr@loc< fun [ v -> match $coercion$ with [ $matchings$ ] ] >>
+      else
+        match body with
+        | `Fun fun_expr ->
+          (* Prevent violation of value restriction and problems with recursive types by
+             eta-expanding function definitions *)
           <:expr@loc< fun [ v -> $fun_expr$ v ] >>
-      | `Match matchings ->
+        | `Match matchings ->
           <:expr@loc< fun [ $matchings$ ] >>
     in
     let patts =
@@ -567,12 +608,8 @@ module Generate_sexp_of = struct
         ~f:(fun ty -> <:patt@loc< $lid:"_of_" ^ Gen.get_tparam_id ty$>>)
     in
     let body = Gen.abstract loc patts body in
-    let body =
-      match mk_poly_type Sig_generate_sexp_of.sig_of_td__loop loc type_name tps with
-      | None -> body
-      | Some typ -> <:expr@loc< ( $body$ : $typ$ ) >>
-    in
-    <:binding@loc< $lid:"sexp_of_" ^ type_name$ = $body$ >>
+    let annot = mk_poly_type Sig_generate_sexp_of.sig_of_td__loop loc type_name tps in
+    <:binding@loc< $lid:"sexp_of_" ^ type_name$ : $annot$ = $body$ >>
 
   let rec sexp_of_tds = function
     | Ast.TyDcl (loc, type_name, tps, rhs, _cl) ->
@@ -597,14 +634,16 @@ module Generate_sexp_of = struct
   let () = Pa_type_conv.add_generator "sexp_of" sexp_of
 
   let sexp_of_exn _rec tp =
-    let get_full_cnstr cnstr = Pa_type_conv.get_conv_path () ^ "." ^ cnstr in
+    let get_full_cnstr str = Pa_type_conv.get_conv_path () ^ "." ^ str in
     let expr =
       match tp with
       | <:ctyp@loc< $uid:cnstr$ >> ->
+          let str = Pa_type_conv.Gen.regular_constr_of_revised_constr cnstr in
           <:expr@loc<
-            Sexplib.Exn_magic.register $uid:cnstr$ $str:get_full_cnstr cnstr$
+            Sexplib.Exn_magic.register $uid:cnstr$ $str:get_full_cnstr str$
           >>
       | <:ctyp@loc< $uid:cnstr$ of $tps$ >> ->
+          let str = Pa_type_conv.Gen.regular_constr_of_revised_constr cnstr in
           let ctyps = Ast.list_of_ctyp tps [] in
           let fps = List.map ~f:sexp_of_type ctyps in
           let sexp_converters =
@@ -626,7 +665,7 @@ module Generate_sexp_of = struct
             let partial =
               <:expr@loc<
                 Sexplib.Exn_magic.$lid:register_name$
-                  $make_exc$ $str:get_full_cnstr cnstr$
+                  $make_exc$ $str:get_full_cnstr str$
               >>
             in
             Gen.apply loc partial sexp_converters
@@ -668,8 +707,9 @@ module Generate_of_sexp = struct
   (* Generate code for matching malformed S-expressions *)
   let mk_variant_other_matches loc rev_els call =
     let coll_structs acc (loc, cnstr) =
+      let str = Pa_type_conv.Gen.regular_constr_of_revised_constr cnstr in
       <:match_case@loc<
-        $str:cnstr$ -> Sexplib.Conv_error.$lid:call$ _tp_loc _sexp
+        $str:str$ -> Sexplib.Conv_error.$lid:call$ _tp_loc _sexp
       >> :: acc
     in
     let exc_no_variant_match =
@@ -721,7 +761,7 @@ module Generate_of_sexp = struct
 
   (* Conversion of types *)
   let rec type_of_sexp = function
-    | <:ctyp@loc< sexp_opaque $_$ >> ->
+    | <:ctyp@loc< sexp_opaque $_$ >> | <:ctyp@loc< _ >> ->
         `Fun <:expr@loc< Sexplib.Conv.opaque_of_sexp >>
     | <:ctyp@loc< sexp_option >> ->
         `Fun <:expr@loc< fun a_of_sexp v -> Some (a_of_sexp v) >>
@@ -785,7 +825,7 @@ module Generate_of_sexp = struct
     let new_other_matches =
       [
         <:match_case@loc<
-          _ -> try ($fun_expr$ _sexp :> $full_type$) with [ $match_exc$ ]
+          _ -> try ($fun_expr$ _sexp :> $replace_variables_by_underscores full_type$) with [ $match_exc$ ]
         >>
       ]
     in
@@ -795,7 +835,8 @@ module Generate_of_sexp = struct
   and mk_variant_match_atom loc full_type rev_atoms_inhs rev_structs =
     let coll (other_matches, match_last) = function
       | `A (loc, cnstr) ->
-          let new_match = <:match_case@loc< $str:cnstr$ -> `$cnstr$ >> in
+          let str = Pa_type_conv.Gen.regular_constr_of_revised_constr cnstr in
+          let new_match = <:match_case@loc< $str:str$ -> `$cnstr$ >> in
           new_match :: other_matches, false
       | `I inh ->
           handle_variant_inh full_type match_last other_matches inh
@@ -855,7 +896,8 @@ module Generate_of_sexp = struct
           has_structs_ref := true;
           let expr = mk_cnstr_args_match ~is_variant:true cnstr tps in
           let new_match =
-            <:match_case@loc< ($str:cnstr$ as _tag) -> $expr$ >>
+            let str = Pa_type_conv.Gen.regular_constr_of_revised_constr cnstr in
+            <:match_case@loc< ($str:str$ as _tag) -> $expr$ >>
           in
           new_match :: other_matches, false
       | `I inh ->
@@ -898,15 +940,8 @@ module Generate_of_sexp = struct
 
   (* Generate matching code for variants *)
   and variant_of_sexp ?full_type row_tp =
-    let rec replace_params_with_underscores = function
-      | <:ctyp@loc< $a$ $_$ >> -> <:ctyp@loc< $replace_params_with_underscores a$ _ >>
-      | x -> x
-    in
     let loc = Ast.loc_of_ctyp row_tp in
     let row_fields = Ast.list_of_ctyp row_tp [] in
-    let row_tp =
-      Ast.tyOr_of_list (List.map row_fields ~f:replace_params_with_underscores)
-    in
     let is_contained, full_type =
       match full_type with
       | None -> true, <:ctyp@loc< [= $row_tp$ ] >>
@@ -917,7 +952,8 @@ module Generate_of_sexp = struct
       | (<:ctyp< $id:_$ >> | <:ctyp< $_$ $_$ >>) as inh :: rest ->
           let rec loop inh row_fields =
             let call =
-              <:expr@loc< ( $mk_internal_call inh$ sexp :> $full_type$ ) >>
+              <:expr@loc< ( $mk_internal_call inh$ sexp :>
+                              $replace_variables_by_underscores full_type$ ) >>
             in
             match row_fields with
             | [] -> call
@@ -978,17 +1014,20 @@ module Generate_of_sexp = struct
   (* Sum type conversions *)
 
   (* Generate matching code for well-formed S-expressions wrt. sum types *)
-  let rec mk_good_sum_matches = function
+  let rec mk_good_sum_matches ctyp =
+    match regular_variant_of_gadt_syntax ctyp with
     | <:ctyp@loc< $uid:cnstr$ >> ->
-        let lccnstr = String.uncapitalize cnstr in
+        let str = Pa_type_conv.Gen.regular_constr_of_revised_constr cnstr in
+        let lcstr = String.uncapitalize str in
         <:match_case@loc<
-          Sexplib.Sexp.Atom ($str:lccnstr$ | $str:cnstr$) -> $uid:cnstr$
+          Sexplib.Sexp.Atom ($str:lcstr$ | $str:str$) -> $uid:cnstr$
         >>
     | <:ctyp@loc< $uid:cnstr$ of $tps$ >> ->
-        let lccnstr = String.uncapitalize cnstr in
+        let str = Pa_type_conv.Gen.regular_constr_of_revised_constr cnstr in
+        let lcstr = String.uncapitalize str in
         <:match_case@loc<
           (Sexplib.Sexp.List
-            [Sexplib.Sexp.Atom ($str:lccnstr$ | $str:cnstr$ as _tag) ::
+            [Sexplib.Sexp.Atom ($str:lcstr$ | $str:str$ as _tag) ::
               sexp_args] as _sexp) ->
                 $mk_cnstr_args_match ~is_variant:false cnstr tps$
         >>
@@ -1001,18 +1040,21 @@ module Generate_of_sexp = struct
 
   (* Generate matching code for malformed S-expressions with good tags
      wrt. sum types *)
-  let rec mk_bad_sum_matches = function
+  let rec mk_bad_sum_matches ctyp =
+    match regular_variant_of_gadt_syntax ctyp with
     | <:ctyp@loc< $uid:cnstr$ >> ->
-        let lccnstr = String.uncapitalize cnstr in
+        let str = Pa_type_conv.Gen.regular_constr_of_revised_constr cnstr in
+        let lcstr = String.uncapitalize str in
         <:match_case@loc<
           Sexplib.Sexp.List
-            [Sexplib.Sexp.Atom ($str:lccnstr$ | $str:cnstr$) :: _] as sexp ->
+            [Sexplib.Sexp.Atom ($str:lcstr$ | $str:str$) :: _] as sexp ->
               Sexplib.Conv_error.stag_no_args _tp_loc sexp
         >>
     | <:ctyp@loc< $uid:cnstr$ of $_$ >> ->
-        let lccnstr = String.uncapitalize cnstr in
+        let str = Pa_type_conv.Gen.regular_constr_of_revised_constr cnstr in
+        let lcstr = String.uncapitalize str in
         <:match_case@loc<
-          Sexplib.Sexp.Atom ($str:lccnstr$ | $str:cnstr$) as sexp ->
+          Sexplib.Sexp.Atom ($str:lcstr$ | $str:str$) as sexp ->
             Sexplib.Conv_error.stag_takes_args _tp_loc sexp
         >>
     | <:ctyp@loc< $tp1$ | $tp2$ >> ->
@@ -1313,17 +1355,19 @@ module Generate_of_sexp = struct
       is_variant_ref := true;
       variant_of_sexp ~full_type row_fields
     in
+    let is_private, rhs = go_through_private_and_manifest_types rhs in
+    if is_private then Loc.raise loc (Failure "of_sexp is not supported for private type");
     let body =
-      let rec loop tp =
-        Gen.switch_tp_def tp
-          ~alias:(fun (_ : Loc.t) tp -> handle_alias tp)
-          ~sum:(fun (_ : Loc.t) tp -> sum_of_sexp tp)
-          ~record:(fun (_ : Loc.t) tp -> record_of_sexp tp)
-          ~variants:(fun (_ : Loc.t) tp -> handle_variant tp)
-          ~mani:(fun (_ : Loc.t) _tp1 tp2 -> loop tp2)
-          ~nil:nil_of_sexp
+      let body =
+        Gen.switch_tp_def rhs
+          ~alias:   (fun (_ : Loc.t) tp  -> handle_alias   tp)
+          ~sum:     (fun (_ : Loc.t) tp  -> sum_of_sexp    tp)
+          ~record:  (fun (_ : Loc.t) tp  -> record_of_sexp tp)
+          ~variants:(fun (_ : Loc.t) tp  -> handle_variant tp)
+          ~mani:    (fun (_ : Loc.t) _ _ -> assert false)
+          ~nil:     nil_of_sexp
       in
-      match loop rhs with
+      match body with
       | `Fun fun_expr ->
           (* Prevent violation of value restriction and problems with
              recursive types by eta-expanding function definitions *)
@@ -1392,18 +1436,11 @@ module Generate_of_sexp = struct
       Gen.abstract loc arg_patts
         <:expr@loc< fun sexp -> ($pre_external_fun_body$) >>
     in
-    let type_of_of_sexp =
-      mk_poly_type Sig_generate_of_sexp.sig_of_td__loop loc type_name tps
-    in
-    let maybe_annotate expr =
-      match type_of_of_sexp with
-      | None -> expr
-      | Some typ -> <:expr@loc< ( $expr$ : $typ$ ) >>
-    in
-    let internal_fun_body = maybe_annotate internal_fun_body in
-    let external_fun_body = maybe_annotate external_fun_body in
-    let internal_binding = <:binding@loc< $lid:internal_name$ = $internal_fun_body$ >> in
-    let external_binding = <:binding@loc< $lid:external_name$ = $external_fun_body$ >> in
+    let annot = mk_poly_type Sig_generate_of_sexp.sig_of_td__loop loc type_name tps in
+    let internal_binding =
+      <:binding@loc< $lid:internal_name$ : $annot$ = $internal_fun_body$ >> in
+    let external_binding =
+      <:binding@loc< $lid:external_name$ : $annot$ = $external_fun_body$ >> in
     internal_binding, external_binding
 
   let rec tds_of_sexp acc = function
